/**
 * Copyright (C) 2007  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.bonita.example.aw;

import java.io.File;
import java.util.Collection;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.security.auth.login.LoginContext;

import org.ow2.bonita.facade.ManagementAPI;
import org.ow2.bonita.facade.QueryRuntimeAPI;
import org.ow2.bonita.facade.RuntimeAPI;
import org.ow2.bonita.facade.def.majorElement.ProcessDefinition;
import org.ow2.bonita.facade.runtime.ActivityInstance;
import org.ow2.bonita.facade.runtime.ActivityState;
import org.ow2.bonita.facade.runtime.ProcessInstance;
import org.ow2.bonita.facade.runtime.TaskInstance;
import org.ow2.bonita.facade.uuid.PackageDefinitionUUID;
import org.ow2.bonita.facade.uuid.ProcessInstanceUUID;
import org.ow2.bonita.facade.uuid.TaskUUID;
import org.ow2.bonita.util.AccessorUtil;
import org.ow2.bonita.util.BonitaConstants;
import org.ow2.bonita.util.BonitaException;
import org.ow2.bonita.util.BonitaRuntimeException;
import org.ow2.bonita.util.Misc;
import org.ow2.bonita.util.SimpleCallbackHandler;
import org.ow2.bonita.util.StandardCallbackHandler;

/**
 * @author "Pierre Vigneras", "Marc Blachon", "Charles Souillard"
 */
public final class ApprovalWorkflow {

  private static final Logger LOG = Logger.getLogger(ApprovalWorkflow.class.getName());
  public static final String PROCESS_ID = "ApprovalWorkflow";

  private static final String BAR_PREFIX = "-bar=";
  private static final String DECISION_PREFIX = "-decision=";
  
  private ApprovalWorkflow() { }

  private static void usage(final String[] args) {
    LOG.severe("Usage: java " + ApprovalWorkflow.class.getName() + " " 
        + BAR_PREFIX + "barFile " + DECISION_PREFIX + "<accepted>");
    LOG.severe("\t<accepted>: whether the request is approved "
        + "('y', 'yes' or 'true') or not (other)");
    LOG.severe("");
    LOG.severe("Actual args:");
    if (args == null) {
      LOG.severe("  args is null");  
    } else {
      for (String arg : args) {
        LOG.severe("  arg:" + arg);
      }
    }
    throw new BonitaRuntimeException("Please read the above usage"); 
  }
  
  public static void main(final String[] args) throws Exception {
    if (args == null || args.length != 2) {
      usage(args);
    }
    byte[] barFile = null;
    String accepted = null;
    for (String arg : args) {
      if (arg.startsWith(BAR_PREFIX)) {
        barFile = Misc.getAllContentFrom(new File(arg.substring(BAR_PREFIX.length())));
      } else if (arg.startsWith(DECISION_PREFIX)) {
        accepted = arg.substring(DECISION_PREFIX.length());
      } else {
        usage(args);
      }
    }
        
    final boolean isGranted = "true".equals(accepted) || "y".equals(accepted) || "yes".equals(accepted);
    
    final String loginMode = System.getProperty(BonitaConstants.LOGIN_MODE_PROPERTY);
    LoginContext loginContext = null;
    if (loginMode != null && BonitaConstants.LOGIN_MODE_TEST.equals(loginMode)) {
      loginContext = new LoginContext("Bonita", new SimpleCallbackHandler("john", "bpm"));
    } else {
      loginContext = new LoginContext("Bonita", new StandardCallbackHandler());
    }
    loginContext.login();
    
    ProcessInstanceUUID instanceUUID = execute(barFile, isGranted);

    cleanPackage(instanceUUID);
    
    loginContext.logout();
    
    if (LOG.isLoggable(Level.INFO)) {
      LOG.info("\n***** End of ApprovalWorkflow Main. *****\n");
    }
  }

  public static ProcessInstanceUUID execute(final byte[] barFile, final boolean isGranted) throws BonitaException {
    //get APIs
    final RuntimeAPI runtimeAPI = AccessorUtil.getAPIAccessor().getRuntimeAPI();
    final QueryRuntimeAPI queryRuntimeAPI = AccessorUtil.getAPIAccessor().getQueryRuntimeAPI();
    final ManagementAPI managementAPI = AccessorUtil.getAPIAccessor().getManagementAPI();

    //deployment
    final ProcessDefinition process = managementAPI.deployBar(barFile).get(PROCESS_ID);
    final PackageDefinitionUUID packageUUID = process.getPackageDefinitionUUID();
    Misc.badStateIfNull(packageUUID, "packageUUID is null in process : " + process.getProcessId());

    //instantiation
    final ProcessInstanceUUID instanceUUID = runtimeAPI.instantiateProcess(process.getUUID());

    //tasks execution
    Collection<ActivityInstance< TaskInstance >> activities = queryRuntimeAPI.getTaskList(instanceUUID, ActivityState.READY);
    if (activities.isEmpty()) {
      throw new BonitaRuntimeException("No task found? Bad User?");
    }
    while (!activities.isEmpty()) {     
      for (ActivityInstance< TaskInstance > activity : activities) {
        final TaskUUID taskUUID = activity.getBody().getUUID();
        final String activityId = activity.getActivityId();
        log("Starting task associated to activity: " + activityId);
        runtimeAPI.startTask(taskUUID, true);
        if (activityId.equals("Approval")) {
          runtimeAPI.setActivityInstanceVariable(activity.getUUID(), "isGranted", isGranted);
        }
        log("Finishing task associated to activity: " + activityId);
        runtimeAPI.finishTask(taskUUID, true);
        log("Task associated to activity: " + activityId + " finished.");
      }
      activities = queryRuntimeAPI.getTaskList(instanceUUID, ActivityState.READY);
    }
    return instanceUUID;
  }
  
  public static void cleanPackage(ProcessInstanceUUID instanceUUID) throws BonitaException {
    final ManagementAPI managementAPI = AccessorUtil.getManagementAPI();
    final QueryRuntimeAPI queryRuntimeAPI = AccessorUtil.getQueryRuntimeAPI();
    final ProcessInstance instance = queryRuntimeAPI.getProcessInstance(instanceUUID);
    final PackageDefinitionUUID packageUUID = instance.getPackageDefinitionUUID();

    //undeployment
    managementAPI.undeploy(packageUUID);
    //journal + history cleaning
    managementAPI.deletePackage(packageUUID);
  }
  
  private static void log(String msg) {
    if (LOG.isLoggable(Level.INFO)) {
      LOG.info(Misc.LINE_SEPARATOR + "***** " + msg + " *****" + Misc.LINE_SEPARATOR);
    }
  }
}
